module Certificate where

import Prelude hiding (product, sum)
import Term
import Text.XML.Light
import INF
import Proof
import CTRS
import TRS
import CRule
import Order
import Precedence
import Status

funapp :: String -> [Content] -> Content
funapp nam content = Elem (unode "funapp" (Elem (unode "name" nam) : content))

var :: String -> Content
var x = Elem (unode "var" x)

rule :: Content -> Content -> Element
rule s t = unode "rule" [s, t]

term :: Term -> Content
term (V x) = var x
term (F f ts) = funapp f (map term ts)

-- TODO: abolish term without ' ???
term' :: Term -> Element
term' (V x) = unode "var" x
term' (F f ts) = unode "funapp" (Elem (name f) : (map (Elem . term') ts))

-- should be [(Term, Term)] -> ... or something?
rules :: [(Content, Content)] -> Element
rules rls = unode "rules" [ rule l r | (l, r) <- rls ]


condition :: (Term, Term) -> Element
condition (s, t) = unode "condition" [ term' s, term' t ] 

conditions :: [(Term, Term)] -> Element
conditions cs = unode "conditions" [ condition c | c <- cs ]

crule :: CRule -> Element
crule (cs, (l, r)) =
  unode "rule" [ term' l, term' r, conditions cs ]

crules :: CTRS -> Element
crules ces = unode "rules" [ crule c | c <- ces ]

trs :: TRS -> Element
trs x = unode "trs" (rules [ (term l, term r) | (l, r) <- x ])

-- NOTE: type is a keyword...
typ :: String -> Element
typ dom =
  unode "type" (unode "polynomial" [unode "domain" (unode dom ()), unode "degree" "1"])

variable :: Int -> Element
variable x = unode "variable" (show x)

integer :: Int -> Element
integer c = unode "integer" (show c)

-- proper naming?
-- c x
-- where c is a coefficient and x is a variable
monomial:: Int -> Int -> Element
monomial c x = unode "product" [integer c, variable x]

name :: String -> Element
name f = unode "name" f

fname :: String -> Content
fname f = Elem (name f)

arity :: String -> Signature -> Content
arity f sig = 
  case lookup f sig of
    Just n -> Elem (unode "arity" (show n))
    Nothing -> error ("function symbol " ++ f ++ " is not found")

interpret :: [Content] -> Element
interpret es = Element (unqual "interpret") [] es Nothing

linear_expression :: (Int, [Int]) -> Content
linear_expression (c, []) = Elem (integer c)
linear_expression (c, cs) =
  Elem (unode "sum" (integer c : [ monomial ci xi | (ci, xi) <- zip cs [1..]  ]))

linear_interpretation :: String -> Signature -> [(String, (Int, [Int]))] -> Element
linear_interpretation dom sig alg =
  unode "interpretation" (typ dom : interps)
  where
    interps = [ interpret [ fname f, Certificate.arity f sig, linear_expression (c0, cs) ] | (f, (c0, cs)) <- alg ]

precedence :: Int -> Content
precedence n = Elem (unode "precedence" (show n))

status :: Status -> String -> Content
status st f =
  case lookup f st of
    Nothing -> error ("function symbol " ++ f ++ " is not found in status " ++ showStatus st)
    Just bs -> Elem (unode "status" [ unode "position" (show i) | (i :: Int, b) <- zip [1..] bs, b ])

precedenceStatus :: Signature -> Precedence -> Status -> Element
precedenceStatus sig prec st =
  unode "precedenceStatus" (do {
    (f, n) <-  shift prec;
    return (unode "precedenceStatusEntry" [
      Elem (name f),
      Certificate.arity f sig,
      Certificate.precedence n,
      status st f
    ])})

nonreachableCorewritePair :: Signature -> Order -> Element
nonreachableCorewritePair sig (PolyNat a) =
  unode "nonreachableCorewritePair" (linear_interpretation "naturals" sig a)
nonreachableCorewritePair sig (PolyInt a) =
  unode "nonreachableCorewritePair" (linear_interpretation "integers" sig a)
nonreachableCorewritePair sig (PolyNegInt a) =
  unode "nonreachableCorewritePair" (linear_interpretation "negativeIntegers" sig a)
nonreachableCorewritePair sig (LinearWPO a p st) =
  unode "nonreachableCorewritePair"
    (unode "weightedPathOrder" [ precedenceStatus sig p st, linear_interpretation "naturals" sig a ])
nonreachableCorewritePair _sig (MaxPlusWPO _a _p _st) = error "to be supported"
nonreachableCorewritePair _sig (LinearSPONat _a _p _st) = error "not supported in CeTA"
nonreachableCorewritePair _sig (LinearSPONegInt _a _p _st) = error "not supported in CeTA"

nonreachabilityProof :: Signature -> Order -> Element
nonreachabilityProof s order =
  (unode "nonreachabilityProof" (nonreachableCorewritePair s order))

infeasibilityProof :: Signature -> Proof -> Content
infeasibilityProof sig (o, (s, t)) = 
  Elem
    (unode "proof"
      (unode "infeasibilityProof"
        (unode "infeasibleEquation"
          [ rule (term s) (term t), nonreachabilityProof sig o])))

conditionType :: INFProblem -> Content
conditionType (Oriented _) =
  Elem (unode "conditionType" (unode "oriented" ()))
-- join and semi-equational are not supported by CeTA, actually
conditionType (Join _) =
  Elem (unode "conditionType" (unode "join" ()))
conditionType (SemiEquational _) =
  Elem (unode "conditionType" (unode "semiEquational" ()))

infeasibilityQuery :: ES -> Content
infeasibilityQuery qs = Elem (unode "infeasibilityQuery" [ rule (term l) (term r) | (l, r) <- qs ])

infeasibilityInput :: INFProblem -> Content
infeasibilityInput p@(Oriented (ces, q)) =
  Elem (unode "infeasibilityInput" [ Elem (unode "ctrs" [ conditionType p, Elem (crules ces) ]), infeasibilityQuery q ])
infeasibilityInput _ = error "CeTA does not support join or semi-equational CTRSs"

metaInfo :: Content
metaInfo =
  Elem (unode "metaInformation" (unode "toolInfos" (unode "toolInfo" "Moca")))

certificate :: INFProblem -> Proof -> String
certificate infp pr =
  showTopElement (Element
    (unqual "certificationProblem")
    [ Attr (unqual "xmlns:xsi") "http://www.w3.org/2001/XMLSchema-instance",
      Attr (QName "noNamespaceSchemaLocation" Nothing (Just "xsi")) "cpf3.xsd"
    ]
    [ Elem (unode "cpfVersion" "3.0"),
      Elem (unode "lookupTables" ()),
      Elem (unode "input" (infeasibilityInput infp)),
      Elem (unode "property" (unode "infeasibility" ())),
      Elem (unode "answer" (unode "yes" ())),
      infeasibilityProof (INF.signatureOf infp) pr
    ]
    Nothing)
