module LinearSPONat where

-- SPO with linear polynomial on N
-- no marking
-- strict precedence (not quasi-)

import Data.List
import Term
import Rule
import CRule
import TRS
import CTRS
import SMT
import Polynomial as P
import Precedence
import Status
import Proof
import Order

type Parameter = (P.Algebra, Precedence, Status)

showParameter :: Parameter -> String
showParameter (a, p, st) =
  showAlgebra a ++
  showPrecedence p ++
  showStatus st

functions :: CTRS -> (Term, Term) -> [String]
functions ctrs q = union (Rule.functions q) (CTRS.functions ctrs)

geqPol :: Polynomial -> Polynomial -> Bool
geqPol v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p x = P.coefficient x v1 >= P.coefficient x v2
 
gtPol :: Polynomial -> Polynomial -> Bool
gtPol v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p "" = P.coefficient "" v1 >  P.coefficient "" v2
    p x  = P.coefficient x  v1 >= P.coefficient x  v2

-- >=_A
geqA :: Algebra -> Term -> Term -> Bool
geqA a l r = LinearSPONat.geqPol (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Term -> Term -> Bool
gtA a l r = LinearSPONat.gtPol (interpret a l) (interpret a r)

-- underlying orders (\sqsupseteq, \sqsupset) resembling WPO
geqSq :: Parameter -> Term -> Term -> Bool
geqSq (a, pr, _) s@(F _ _) t@(F _ _) =
  gtA a (sharp s) (sharp t) || (geqA a (sharp s) (sharp t) && geqPrec pr s t)
geqSq _ s t = s == t

gtSq :: Parameter -> Term -> Term -> Bool
gtSq (a, pr, _) s@(F _ _) t@(F _ _) =
  gtA a (sharp s) (sharp t) || (geqA a (sharp s) (sharp t) && gtPrec pr s t)
gtSq _ _ _ = False

geqSPO :: Parameter -> Term -> Term -> Bool
geqSPO _ (V x) (V y) = x == y
geqSPO _ (V _) (F _ _) = False
geqSPO param@(_, _, st) s@(F _ _) t@(V _) =
  any (\s' -> geqSPO param s' t) (args st s)
geqSPO param@(_, _, st) s@(F _ _) t@(F _ _) = 
  any (\s' -> geqSPO param s' t) (args st s) ||
  (all (\t' -> gtSPO param s t') (args st t) &&
    (gtSq param s t || (geqSq param s t && geqSPOLex param (args st s) (args st t))))

gtSPO :: Parameter -> Term -> Term -> Bool
gtSPO _ (V _) _ = False
gtSPO param@(_, _, st) s@(F _ _) t@(V _) =
  any (\s' -> geqSPO param s' t) (args st s)
gtSPO param@(_, _, st) s@(F _ _) t@(F _ _) = 
  any (\s' -> geqSPO param s' t) (args st s) ||
  (all (\t' -> gtSPO param s t') (args st t) &&
    (gtSq param s t || (geqSq param s t && gtSPOLex param (args st s) (args st t))))

-- args are assumed to be already filtered
geqSPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqSPOLex _ [] [] = True
geqSPOLex _ [] (_ : _) = False
geqSPOLex _ (_ : _) []  = True
geqSPOLex param (s : ss) (t : ts) =
  gtSPO param s t || 
  (geqSPO param s t && geqSPOLex param ss ts)

gtSPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtSPOLex _ [] [] = False
gtSPOLex _ [] (_ : _) = False
gtSPOLex _ (_ : _) []  = True
gtSPOLex param (s : ss) (t : ts) =
  gtSPO param s t || 
  (geqSPO param s t && gtSPOLex param ss ts)

geqMSPO :: Parameter -> Term -> Term -> Bool
geqMSPO param@(a, _, _) s t = geqA a s t && geqSPO param s t

-- Encoding functions.

type Table = [(Term,Term)]

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqSPO :: Table -> Rule -> Formula
varGeqSPO table rule
  | Just i <- elemIndex rule table = FVar ("geqSPO_" ++ show i)
  | otherwise                      = error "varGeqSPO"

varGtSPO :: Table -> Rule -> Formula
varGtSPO table rule
  | Just i <- elemIndex rule table = FVar ("gtSPO_" ++ show i)
  | otherwise                      = error "varGtSPO"

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (LinearSPONat.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (LinearSPONat.coefficient x l) (LinearSPONat.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (LinearSPONat.coefficient x l) (LinearSPONat.coefficient x r) 
        | x <- Rule.variables rule ])

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

encodeGeqSq :: Rule -> Formula
encodeGeqSq (s@(F _ _), t@(F _ _)) =
  disj [ encodeGtA (sharp s, sharp t), conj [ encodeGeqA (sharp s, sharp t), encodeGeqPrec (s, t) ] ]
encodeGeqSq (s, t)
  | s == t = top
  | otherwise = bottom

encodeGtSq :: Rule -> Formula
encodeGtSq (s@(F _ _), t@(F _ _)) =
  disj [ encodeGtA (sharp s, sharp t), conj [ encodeGeqA (sharp s, sharp t), encodeGtPrec (s, t) ] ]
encodeGtSq _ = bottom

-- s >=_spo t
encodeGeqSPO :: Table -> Rule -> Formula
encodeGeqSPO _ (s@(V _), t)
  | s == t = top
  | otherwise = bottom
encodeGeqSPO table rl@(F _ _, V _) = encodeSPOSub table rl
encodeGeqSPO table rl@(F f ss, F g ts) =
  disj [
    encodeSPOSub table rl,
    conj [
      encodeSPOArgs table rl,
      disj [
        encodeGtSq rl,
        -- only strict precedence is considered for now
        if f == g then conj [encodeGeqSq rl, encodeGeqSPOLex table f ss ts] else bottom
      ]
    ]
  ]

-- s >_spo t
encodeGtSPO :: Table -> Rule -> Formula
encodeGtSPO _ (V _, _) = bottom
encodeGtSPO table rl@(F _ _, V _) = encodeSPOSub table rl
encodeGtSPO table rl@(F f ss, F g ts) =
  disj [
    encodeSPOSub table rl,
    conj [
      encodeSPOArgs table rl,
      disj [
        encodeGtSq rl,
        -- only strict precedence is considered for now
        if f == g then conj [encodeGeqSq rl, encodeGtSPOLex table f ss ts] else bottom
      ]
    ]
  ]

encodeSPOSub :: Table -> Rule -> Formula
encodeSPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqSPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeSPOSub _ _ = error "encodeSPOSub"

encodeSPOArgs :: Table -> Rule -> Formula
encodeSPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtSPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeSPOArgs _ _ = error "encodeSPOArgs"

encodeGtSPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtSPOLex tbl f ss ts = encodeGtSPOLex' tbl f 1 ss ts

encodeGtSPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtSPOLex' _ _ _ [] [] = bottom
encodeGtSPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtSPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqSPO tbl (s, t)), encodeGtSPOLex' tbl f (i+1) ss ts ] ]
encodeGtSPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqSPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqSPOLex tbl f ss ts = encodeGeqSPOLex' tbl f 1 ss ts

encodeGeqSPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqSPOLex' _ _ _ [] [] = top
encodeGeqSPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtSPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqSPO tbl (s, t)), encodeGeqSPOLex' tbl f (i+1) ss ts ] ]
encodeGeqSPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqMSPO :: Table -> Rule -> Formula
encodeGeqMSPO tbl rl = conj [encodeGeqA rl, encodeGeqSPO tbl rl ]

encodeSPO :: CTRS -> [(Term, Term)] -> Formula
encodeSPO ctrs qs =
  conj (disj [ varGtSPO table (t, s) | (s, t) <- qs ] :
        [ encodeGeqMSPO table rule | (_, rule) <- ctrs ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varF0 (f ++ "#")) (Val 0) | (f, _) <- fs ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtSPO table rule) (encodeGtSPO table rule) | rule <- table ] ++
        [ implies (varGeqSPO table rule) (encodeGeqSPO table rule) | rule <- table ])
  where 
    fs = union (CTRS.signatureOf ctrs) (TRS.signatureOf qs)
    table = nub ([ (s, t) | (l, r) <- (ctrs2trs ctrs), s <- subterms l, t <- subterms r ] ++ 
                 [ (s, t) | (l, r) <- qs, s <- subterms r, t <- subterms l ])

-- -- Decoding functions

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: CTRS -> [(Term, Term)] -> Model -> Parameter
decode ctrs qs model =
  (decodeAlgebra (sig ++ sharped) model,
   decodePrecedence sig model,
   decodeStatus sig model)
  where
    sig =  union (CTRS.signatureOf ctrs) (TRS.signatureOf qs)
    sharped = [ (f ++ "#", n) | (f, n) <- sig ]

checkCRule :: Parameter -> CRule -> IO ()
checkCRule p rule@(_cs, (l, r)) = 
  if geqMSPO p l r
    then return ()
    else error ("checkCRule: " ++ showCRule rule ++ " is not oriented.\n" ++ showParameter p ++ "\n")

checkCTRS :: Parameter -> CTRS -> IO ()
checkCTRS p ctrs = mapM_ (checkCRule p) ctrs

checkQuery :: Parameter -> [(Term, Term)] -> IO (Term, Term)
checkQuery p [] = error ("checkQuery: no query is oriented.\n" ++ showParameter p ++ "\n")
checkQuery p ((s, t) : qs) =  
  if gtSPO p t s
    then return (s, t)
    else checkQuery p qs

solve :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
solve tool ctrs qs = do
  result <- sat tool (encodeSPO ctrs qs)
  case result of
    Nothing    -> return Nothing
    Just model -> do
      let param@(a, p, st) = decode ctrs qs model
      checkCTRS param ctrs
      q <- checkQuery param qs
      return (Just (LinearSPONat a p st, q))
