module LinearWPO where

-- WPO with linear polynomial on N
-- strict precedence (not quasi-)

import Data.List
import Term
import Rule
import CRule
import TRS
import CTRS
import SMT
import Polynomial as P
import Precedence
import Status
import Proof
import Order

type Parameter = (P.Algebra, Precedence, Status)

showParameter :: Parameter -> String
showParameter (a, p, st) =
  showAlgebra a ++
  showPrecedence p ++
  showStatus st

functions :: CTRS -> (Term, Term) -> [String]
functions ctrs q = union (Rule.functions q) (CTRS.functions ctrs)

geqPol :: Polynomial -> Polynomial -> Bool
geqPol v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p x = P.coefficient x v1 >= P.coefficient x v2
 
gtPol :: Polynomial -> Polynomial -> Bool
gtPol v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p "" = P.coefficient "" v1 >  P.coefficient "" v2
    p x  = P.coefficient x  v1 >= P.coefficient x  v2

-- >=_A
geqA :: Algebra -> Term -> Term -> Bool
geqA a l r = LinearWPO.geqPol (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Term -> Term -> Bool
gtA a l r = LinearWPO.gtPol (interpret a l) (interpret a r)

geqWPO :: Parameter -> Term -> Term -> Bool
geqWPO (_, _, _) (V x) (V y) = x == y
geqWPO (a, _, _) s@(V _) t@(F _ _) = gtA a s t
geqWPO param@(a, _, st) s@(F _ _) t@(V _) =
  gtA a s t || (geqA a s t && any (\s' -> geqWPO param s' t) (args st s))
geqWPO param@(a, p, st) s@(F _ _) t@(F _ _) = 
  gtA a s t ||
    (geqA a s t &&
      (any (\s' -> geqWPO param s' t) (args st s) ||
       (all (\t' -> gtWPO param s t') (args st t) &&
        (gtPrec p s t || (geqPrec p s t && geqWPOLex param (args st s) (args st t))))))

gtWPO :: Parameter -> Term -> Term -> Bool
gtWPO (a, _, _) s@(V _) t = gtA a s t
gtWPO param@(a, _, st) s@(F _ _) t@(V _) =
  gtA a s t || (geqA a s t && any (\s' -> geqWPO param s' t) (args st s))
gtWPO param@(a, p, st) s@(F _ _) t@(F _ _) = 
  gtA a s t ||
    (geqA a s t &&
      (any (\s' -> geqWPO param s' t) (args st s) ||
       (all (\t' -> gtWPO param s t') (args st t) &&
        (gtPrec p s t || (geqPrec p s t && gtWPOLex param (args st s) (args st t))))))

-- args are assumed to be already filtered
geqWPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqWPOLex _ [] [] = True
geqWPOLex _ [] (_ : _) = False
geqWPOLex _ (_ : _) []  = True
geqWPOLex param (s : ss) (t : ts) =
  gtWPO param s t || 
  (geqWPO param s t && geqWPOLex param ss ts)

gtWPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtWPOLex _ [] [] = False
gtWPOLex _ [] (_ : _) = False
gtWPOLex _ (_ : _) []  = True
gtWPOLex param (s : ss) (t : ts) =
  gtWPO param s t || 
  (geqWPO param s t && gtWPOLex param ss ts)

-- Encoding functions.

type Table = [(Term,Term)]

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqWPO :: Table -> Rule -> Formula
varGeqWPO table rule
  | Just i <- elemIndex rule table = FVar ("geqWPO_" ++ show i)
  | otherwise                      = error "varGeqWPO"

varGtWPO :: Table -> Rule -> Formula
varGtWPO table rule
  | Just i <- elemIndex rule table = FVar ("gtWPO_" ++ show i)
  | otherwise                      = error "varGtWPO"

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (LinearWPO.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (LinearWPO.coefficient x l) (LinearWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (LinearWPO.coefficient x l) (LinearWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

-- s >_wpo t
encodeGtWPO :: Table -> Rule -> Formula
encodeGtWPO _ rl@(V _, _) = encodeGtA rl
encodeGtWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeWPOSub table rl] ]
encodeGtWPO table rl@(F f ss, F g ts) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeWPOSub table rl,
        conj [
          encodeWPOArgs table rl,
          disj [
            encodeGtPrec rl,
            -- only strict precedence is considered for now
            if f == g then encodeGtWPOLex table f ss ts else bottom
          ]
        ]
      ]
    ]
  ]


-- s >=_wpo t
encodeGeqWPO :: Table -> Rule -> Formula
encodeGeqWPO _ (V x, V y) = if x == y then top else bottom
encodeGeqWPO _ rl@(V _, _) = encodeGtA rl
encodeGeqWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeWPOSub table rl] ]
encodeGeqWPO table rl@(F f ss, F g ts) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeWPOSub table rl,
        conj [
          encodeWPOArgs table rl,
          disj [
            encodeGtPrec rl,
            -- only strict precedence is considered for now
            if f == g then encodeGeqWPOLex table f ss ts else bottom
          ]
        ]
      ]
    ]
  ]

encodeWPOSub :: Table -> Rule -> Formula
encodeWPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqWPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeWPOSub _ _ = error "encodeWPOSub"

encodeWPOArgs :: Table -> Rule -> Formula
encodeWPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtWPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeWPOArgs _ _ = error "encodeWPOArgs"

encodeGtWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtWPOLex tbl f ss ts = encodeGtWPOLex' tbl f 1 ss ts

encodeGtWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtWPOLex' _ _ _ [] [] = bottom
encodeGtWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqWPO tbl (s, t)), encodeGtWPOLex' tbl f (i+1) ss ts ] ]
encodeGtWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqWPOLex tbl f ss ts = encodeGeqWPOLex' tbl f 1 ss ts

encodeGeqWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqWPOLex' _ _ _ [] [] = top
encodeGeqWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqWPO tbl (s, t)), encodeGeqWPOLex' tbl f (i+1) ss ts ] ]
encodeGeqWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeWPO :: CTRS -> [(Term, Term)] -> Formula
encodeWPO ctrs qs =
  conj (disj [ varGtWPO table (t, s) | (s, t) <- qs ] :
        [ varGeqWPO table rule | (_, rule) <- ctrs ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtWPO table rule) (encodeGtWPO table rule) | rule <- table ] ++
        [ implies (varGeqWPO table rule) (encodeGeqWPO table rule) | rule <- table ])
  where 
    fs = union (CTRS.signatureOf ctrs) (TRS.signatureOf qs)
    table = nub ([ (s, t) | (l, r) <- (ctrs2trs ctrs), s <- subterms l, t <- subterms r ] ++ 
                 [ (s, t) | (l, r) <- qs, s <- subterms r, t <- subterms l ])

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: CTRS -> [(Term, Term)] -> Model -> Parameter
decode ctrs qs model =
  (decodeAlgebra sig model,
   decodePrecedence sig model,
   decodeStatus sig model)
  where sig =  union (CTRS.signatureOf ctrs) (TRS.signatureOf qs)

checkCRule :: Parameter -> CRule -> IO ()
checkCRule p rule@(_cs, (l, r)) = 
  if geqWPO p l r
    then return ()
    else error ("checkCRule: " ++ showCRule rule ++ " is not oriented.\n" ++ showParameter p ++ "\n")

checkCTRS :: Parameter -> CTRS -> IO ()
checkCTRS p ctrs = mapM_ (checkCRule p) ctrs

checkQuery :: Parameter -> [(Term, Term)] -> IO (Term, Term)
checkQuery p [] = error ("checkQuery: no query is oriented.\n" ++ showParameter p ++ "\n")
checkQuery p ((s, t) : qs) =  
  if gtWPO p t s
    then return (s, t)
    else checkQuery p qs

solve ::
  String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
solve tool ctrs qs = do
  result <- sat tool (encodeWPO ctrs qs)
  case result of
    Nothing    -> return Nothing
    Just model -> do
      let param@(a, p, st) = decode ctrs qs model
      checkCTRS param ctrs
      q <- checkQuery param qs
      return (Just (LinearWPO a p st, q))
