module Main where

import Options.Applicative

import Term
import CTRS
import ARIParser
import INF
import qualified PolyNat as PN
import qualified PolyInt as PI
import qualified PolyNegInt as PNI
import Proof
import Certificate
import LinearWPO
import MaxPlusWPO
import LinearSPONat
import LinearSPONegInt

data Command = Command
  { order :: String,
    file :: String, 
    cert :: Bool }

opt :: Parser Command
opt = Command
  <$> strOption
    ( long "order"
    <> short 'o'
    <> metavar "ORDER"
    <> help "term ordering (options: poly-nat, poly-int, poly-negint, wpol, wpomp, lpo, spol-nat, spol-negint)")
  <*> strOption
    ( long "file"
    <> short 'f'
    <> metavar "TARGET"
    <> help "target file in ARI format")
  <*> switch
    ( long "cert"
    <> short 'c'
    <> help "generate a certificate in CPF-3")

main :: IO ()
main = dispatch =<< execParser opts
  where
    opts = info (opt <**> helper)
      ( fullDesc <> header "inf-tool - a tool for infeasibility" )

dispatch :: Command -> IO ()
dispatch (Command o f c) =
  case o of
    "poly-nat" -> inf (Main.polyNat tool) f c
    "poly-int" -> inf (Main.polyInt tool) f c
    "poly-negint" -> inf (Main.polyNegInt tool) f c
    "wpol" -> inf (Main.wpol tool) f c
    "wpomp" -> inf (Main.wpomp tool) f c
    "lpo" -> inf (Main.lpo tool) f c
    "spol-nat" -> inf (Main.spolNat tool) f c
    "spol-negint" -> inf (Main.spolNegInt tool) f c
    _ -> error ("order " ++ o ++ " is not supported")
  where tool = "z3"

inf :: (CTRS -> [(Term, Term)] -> IO (Maybe Proof)) -> String -> Bool -> IO ()
inf method filename c = do
  errorOrTRS <- readARI filename
  case errorOrTRS of
    Left e    -> error (show e)
    Right i -> do
      case i of
        Oriented (ctrs, qs) -> do
          result <- method ctrs qs
          case result of
            Nothing -> putStrLn "MAYBE"
            Just p -> report i p
        _ -> do
              putStrLn "MAYBE"
              putStrLn "only oriented (C)TRSs are supported."
  where
    report infp pr =
      if c
        then do
          putStrLn "YES" 
          putStrLn (certificate infp pr)
        else do 
          putStrLn "YES"
          putStrLn ""
          putStr   (showProof pr)

polyInt :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
polyInt tool ctrs qs = do
  m <- PI.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

polyNegInt :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
polyNegInt tool ctrs qs = do
  m <- PNI.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

polyNat :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
polyNat tool ctrs qs = do
  m <- PN.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

wpol :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
wpol tool ctrs qs = do
  m <- LinearWPO.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

wpomp :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
wpomp tool ctrs qs = do
  m <- MaxPlusWPO.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

lpo :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
lpo tool ctrs qs = do
  m <- MaxPlusWPO.solve_lpo tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

spolNat :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
spolNat tool ctrs qs = do
  m <- LinearSPONat.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing

spolNegInt :: String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
spolNegInt tool ctrs qs = do
  m <- LinearSPONegInt.solve tool ctrs qs
  case m of
    Just p -> return (Just p)
    Nothing -> return Nothing
