module MaxPlus where

import Data.List
import Term
import Polynomial as P

-- poly

geqPol :: Polynomial -> Polynomial -> Bool
geqPol v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p x = P.coefficient x v1 >= P.coefficient x v2
 
gtPol :: Polynomial -> Polynomial -> Bool
gtPol v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p "" = P.coefficient "" v1 >  P.coefficient "" v2
    p x  = P.coefficient x  v1 >= P.coefficient x  v2

-- max/plus

-- (f0, [(f1, f'1), ..., (fn, f'n)]) represents
-- max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type MaxPlusParam = (Int, [(Int, Int)])

-- f(x1,...,xn) = max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type Interpretation = (String, MaxPlusParam)

type Algebra = [MaxPlus.Interpretation]

-- pretty printer for f + f'x
showLinearPolynomial :: (Int, Int, String) -> String
showLinearPolynomial (f, 0, _x) = show f
showLinearPolynomial (0, 1, x) = x
showLinearPolynomial (f, 1, x) = show f ++ " + " ++ x
showLinearPolynomial (0, f', x) = show f' ++ x
showLinearPolynomial (f, f', x) = show f ++ " + " ++ show f' ++ " " ++ x

-- can be improved
showMaxPlus :: MaxPlusParam -> String
showMaxPlus (a0, []) = show a0
showMaxPlus (a0, as) = "max{" ++ (intercalate ", " ms) ++ "}"
  where
    m1n = [ showLinearPolynomial (ai, ai', "x" ++ show i)
          | (i, (ai, ai')) <- zip [ 1 :: Int ..] as]
    ms = show a0 : m1n

showInterpretation :: MaxPlus.Interpretation -> String
showInterpretation (f, (a0, [])) =
  f ++ "_A = " ++ show a0
showInterpretation (f, mp@(_, as)) =
  f ++ "_A(" ++ xs ++ ") = " ++ showMaxPlus mp
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: MaxPlus.Algebra -> String
showAlgebra interpretations =
  "max/plus interpretations on N:\n" ++
  "\n" ++
  unlines 
    [ "  " ++ MaxPlus.showInterpretation interpretation 
    | interpretation <- interpretations ] ++ "\n"

-- max{p1, ..., pn}
type MaxPlus = [P.Polynomial]

var :: String -> MaxPlus
var x = [ P.var x ]

num :: Int -> MaxPlus
num n = [ P.num n ]

-- unfold max
flatten :: [MaxPlus] -> MaxPlus
flatten mps = concat mps

plus :: Int -> MaxPlus -> MaxPlus
plus k ps = [P.plus [(P.num k), p] | p <- ps]

times :: Int -> MaxPlus -> MaxPlus
times k ps = [ P.times k p | p <- ps]

-- max G >= max H
mpGeq :: MaxPlus -> MaxPlus -> Bool
mpGeq gs hs = all (\h -> (any (\g -> geqPol g h) gs)) hs

-- max G > max H
mpGt :: MaxPlus -> MaxPlus -> Bool
mpGt gs hs = all (\h -> (any (\g -> gtPol g h) gs)) hs

interpret :: MaxPlus.Algebra -> Term -> MaxPlus
interpret _ (V x) = MaxPlus.var x 
interpret a (F f ts)
  | Just (a0, as) <- lookup f a = 
    MaxPlus.flatten
      (MaxPlus.num a0 :
      [ MaxPlus.plus ai (MaxPlus.times ai' (MaxPlus.interpret a ti)) | ((ai, ai'), ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: MaxPlus.Algebra -> Term -> Term -> Bool
geqA a l r = mpGeq (MaxPlus.interpret a l) (MaxPlus.interpret a r)

-- >_A
gtA :: MaxPlus.Algebra -> Term -> Term -> Bool
gtA a l r = mpGt (MaxPlus.interpret a l) (MaxPlus.interpret a r)

