module PolyNegInt where

import Data.List

import Term
import Rule
import CRule
import TRS
import CTRS
import SMT
import Polynomial as P
import Proof
import Order

functions :: CTRS -> (Term, Term) -> [String]
functions ctrs q = union (Rule.functions q) (CTRS.functions ctrs)

geq :: Polynomial -> Polynomial -> Bool
geq v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p "" = P.coefficient "" v1 >= P.coefficient "" v2
    p x  = P.coefficient x  v1 <= P.coefficient x  v2
 
gt :: Polynomial -> Polynomial -> Bool
gt v1 v2 = all p (union (P.variables v1) (P.variables v2))
  where
    p "" = P.coefficient "" v1 >  P.coefficient "" v2
    p x  = P.coefficient x  v1 <= P.coefficient x  v2

-- >=_A
geqA :: Algebra -> Term -> Term -> Bool
geqA a l r = PolyNegInt.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Term -> Term -> Bool
gtA a l r = PolyNegInt.gt (interpret a l) (interpret a r)

-- Encoding functions.

varF0 :: String -> Exp
varF0 f = Var ("a0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (PolyNegInt.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (PolyNegInt.coefficient x r) (PolyNegInt.coefficient x l) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (PolyNegInt.coefficient x r) (PolyNegInt.coefficient x l) 
        | x <- Rule.variables rule ])

cruleConstraint :: CRule -> Formula
cruleConstraint (_cs, (l, r)) = encodeGeqA (l, r)

encodePolyNegInt :: CTRS -> [(Term, Term)] -> Formula
encodePolyNegInt ctrs qs =
  conj (disj [ encodeGtA (t, s) | (s, t) <- qs ]  :
        [ cruleConstraint crule | crule <- ctrs] ++
        [ SMT.geq (Val 0) (varF0 f) | f <- fs])
  where 
    fs = union (TRS.functions qs) (CTRS.functions ctrs)

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> P.Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decode :: CTRS -> [(Term, Term)] -> Model -> P.Algebra
decode ctrs qs model =
  [ decodeInterpretation f n model 
  | (f, n) <- union (CTRS.signatureOf ctrs) (TRS.signatureOf qs)]

checkCRule :: P.Algebra -> CRule -> IO ()
checkCRule p rule@(_cs, (l, r)) = 
  if PolyNegInt.geqA p l r
    then return ()
    else error ("checkCRule: " ++ showCRule rule ++ " is not oriented.\n" ++ P.showAlgebra p ++ "\n")

checkCTRS :: P.Algebra -> CTRS -> IO ()
checkCTRS a ctrs = mapM_ (checkCRule a) ctrs

checkQuery :: P.Algebra -> [(Term, Term)] -> IO (Term, Term)
checkQuery p [] = error ("checkQuery: no query is not oriented.\n" ++ P.showAlgebra p ++ "\n")
checkQuery p ((s, t) : qs) =  
  if PolyNegInt.gtA p t s
    then return (s, t)
    else checkQuery p qs

solve ::
  String -> CTRS -> [(Term, Term)] -> IO (Maybe Proof)
solve tool ctrs qs = do
  result <- sat tool (encodePolyNegInt ctrs qs)
  case result of
    Nothing    -> return Nothing
    Just model -> do
      let param = decode ctrs qs model
      checkCTRS param ctrs
      q <- checkQuery param qs
      return (Just (PolyNegInt param, q))
